""" Program for validating that simulation tools are consistent with the BioSimulators standards

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-22
:Copyright: 2020, BioSimulators Team
:License: MIT
"""

from .data_model import OutputMedium
from .results.data_model import TestCaseResultType
from .results.io import write_test_results
from biosimulators_utils.config import Colors
import biosimulators_test_suite
import biosimulators_test_suite.exec_core
import cement
import termcolor


class BaseController(cement.Controller):
    """ Base controller for command line application """

    class Meta:
        label = 'base'
        description = "Validates that a simulation tool is consistent with the BioSimulators standards"
        help = "Validates that a simulation tool is consistent with the BioSimulators standards"
        arguments = [
            (['specifications'], dict(
                type=str,
                help='Path or URL to the specifications of the simulator',
            )),
            (['-c', '--test-case'], dict(
                type=str,
                nargs='+',
                default=None,
                dest='case_ids',
                help=(
                    "Ids of test cases of evaluate (e.g., "
                      "'sedml.SimulatorSupportsModelAttributeChanges', "
                      "'published_project.SimulatorCanExecutePublishedProject:"
                      "sbml-core/Caravagna-J-Theor-Biol-2010-tumor-suppressive-oscillations'). "
                      "Default: evaluate all test cases"
                ),
            )),
            (['--cli'], dict(
                default=None,
                help=("Command-line interface to use to execute the tests involving "
                      "the simulation of COMBINE/OMEX archives rather than a Docker image"),
            )),
            (['--synthetic-archives-dir'], dict(
                default=None,
                help="Directory to save the synthetic COMBINE/OMEX archives generated by the test cases",
            )),
            (['--report'], dict(
                default=None,
                help="Path to save a report of the results in JSON format",
            )),
            (['--verbose'], dict(
                action='store_true',
                help="If set, print the stdout and stderr of the execution of the tests in real time.",
            )),
            (['--work-dir'], dict(
                default=None,
                help=(
                    "Working directory for files for evaluating tests. This option enables intermediate files "
                    "to be inspected rather than automatically cleaned up."
                ),
            )),
            (['--do-not-validate-specs'], dict(
                action='store_true',
                help="If set, don't validate the specifications of the simulator.",
            )),
            (['--do-not-log-std-out-err'], dict(
                action='store_true',
                help="If set, don't use capturer to collect stdout and stderr.",
            )),
            (['--dry-run'], dict(
                action='store_true',
                help="If set, create synthetic archives, but do not use the simulator to execute them.",
            )),
            (['-v', '--version'], dict(
                action='version',
                version=biosimulators_test_suite.__version__,
            )),
        ]

    @cement.ex(hide=True)
    def _default(self):
        args = self.app.pargs
        try:
            # execute tests
            validator = biosimulators_test_suite.exec_core.SimulatorValidator(
                args.specifications,
                case_ids=args.case_ids,
                verbose=args.verbose,
                synthetic_archives_dir=args.synthetic_archives_dir,
                output_medium=OutputMedium.console,
                log_std_out_err=not args.do_not_log_std_out_err,
                working_dirname=args.work_dir,
                dry_run=args.dry_run,
                cli=args.cli,
                validate_specs=not args.do_not_validate_specs)
            results = validator.run()

            # print summary
            summary, failure_details, warning_details, skipped_details = validator.summarize_results(results, debug=args.debug)
            print('')
            print('=============== SUMMARY ===============')
            print('')
            print(summary + '\n\n')
            if failure_details:
                color = Colors.failure.value
                print(termcolor.colored('=============== FAILURES ===============', color))
                print(termcolor.colored('', color))
                print(termcolor.colored('* ' + '\n\n* '.join(failure_details), color))
                print('')
            if warning_details:
                color = Colors.warning.value
                print(termcolor.colored('=============== WARNINGS ===============', color))
                print(termcolor.colored('', color))
                print(termcolor.colored('* ' + '\n\n* '.join(warning_details), color))
                print('')
            if skipped_details:
                color = Colors.skipped.value
                print(termcolor.colored('================ SKIPS =================', color))
                print(termcolor.colored('', color))
                print(termcolor.colored('* ' + '\n\n* '.join(skipped_details), color))
                print('')

            # optionally, save report of results to a JSON file
            if args.report:
                write_test_results(results, args.report)
        except Exception as exception:
            raise SystemExit(str(exception))

        any_passed = False
        failed = False
        for result in results:
            if result.type == TestCaseResultType.passed:
                any_passed = True
            elif result.type == TestCaseResultType.failed:
                failed = True

        if failed:
            exit(1)
        if not any_passed:
            exit(3)


class App(cement.App):
    """ Command line application """
    class Meta:
        label = 'biosimulators-test-suite'
        base_controller = 'base'
        handlers = [
            BaseController,
        ]


def main():
    with App() as app:
        app.run()



""" Methods for test cases involving checking support for the COMBINE/OMEX standards

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import InvalidOutputsWarning
from .published_project import SingleMasterSedDocumentCombineArchiveTestCase
from biosimulators_utils.combine.data_model import CombineArchive, CombineArchiveContent, CombineArchiveContentFormat  # noqa: F401
from biosimulators_utils.report.io import ReportReader
from biosimulators_utils.sedml.data_model import SedDocument, Report  # noqa: F401
import abc
import copy
import os
import shutil
import warnings

__all__ = [
    'CombineArchiveTestCase',
    'WhenACombineArchiveHasAMasterFileSimulatorOnlyExecutesThisFile',
    'WhenACombineArchiveHasNoMasterFileSimulatorExecutesAllSedDocuments',
    'CombineArchiveHasSedDocumentsInNestedDirectories',
    'CombineArchiveHasSedDocumentsWithSameNamesInDifferentInNestedDirectories',
]


class CombineArchiveTestCase(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Base class for testing for the execution of master and non-master files in COMBINE/OMEX archives

    Attributes:
        _expected_report_ids (:obj:`list` of :obj:`str`): ids of expected reports
    """

    SED_DOCUMENT_LOCATIONS_ARE_NESTED = False
    SED_DOCUMENTS_HAVE_SAME_NAMES = False

    @abc.abstractmethod
    def _is_concrete():
        pass  # pragma: no cover

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(CombineArchiveTestCase, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        archive = expected_results_of_synthetic_archives[0].archive
        sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        if self.SED_DOCUMENT_LOCATIONS_ARE_NESTED:
            for content in list(archive.contents):
                location_b = os.path.join('subdir', os.path.relpath(content.location, '.'))
                if self.SED_DOCUMENTS_HAVE_SAME_NAMES:
                    location_b_copy = os.path.join('subdir__copy', os.path.relpath(content.location, '.'))

                if not os.path.isdir(os.path.dirname(os.path.join(curated_archive_dir, location_b))):
                    os.makedirs(os.path.dirname(os.path.join(curated_archive_dir, location_b)))
                    if self.SED_DOCUMENTS_HAVE_SAME_NAMES:
                        os.makedirs(os.path.dirname(os.path.join(curated_archive_dir, location_b_copy)))

                shutil.copy(os.path.join(curated_archive_dir, content.location), os.path.join(curated_archive_dir, location_b))
                if self.SED_DOCUMENTS_HAVE_SAME_NAMES:
                    shutil.copy(os.path.join(curated_archive_dir, content.location), os.path.join(curated_archive_dir, location_b_copy))
                if content.format == CombineArchiveContentFormat.SED_ML:
                    sed_docs[location_b] = sed_docs.pop(content.location)
                content.location = location_b
                if content.format != CombineArchiveContentFormat.SED_ML and self.SED_DOCUMENTS_HAVE_SAME_NAMES:
                    content_copy = copy.deepcopy(content)
                    content_copy.location = location_b_copy
                    archive.contents.append(content_copy)

        doc_location = list(sed_docs.keys())[0]
        doc = sed_docs[doc_location]

        if not self.SED_DOCUMENTS_HAVE_SAME_NAMES:
            name, ext = os.path.splitext(doc_location)
            doc_location_2 = name + '__copy' + ext
        elif self.SED_DOCUMENT_LOCATIONS_ARE_NESTED:
            doc_location_2 = os.path.join('subdir__copy', *os.path.split(doc_location)[1:])

        doc_2 = copy.deepcopy(doc)
        sed_docs[doc_location_2] = doc_2
        archive.contents.append(
            CombineArchiveContent(
                doc_location_2,
                CombineArchiveContentFormat.SED_ML,
                master=False),
        )

        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        try:
            report_ids = ReportReader().get_ids(outputs_dir)
        except Exception:
            report_ids = []

        expected_report_ids = self.get_expected_reports(synthetic_archive, synthetic_sed_docs)

        missing_reports = set(expected_report_ids).difference(set(report_ids))
        unexpected_reports = set(report_ids).difference(set(expected_report_ids))

        if missing_reports:
            raise ValueError('Simulator did not generate the following expected reports\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_reports))))

        if unexpected_reports:
            warnings.warn('Simulator generated the following extra unexpected reports:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in unexpected_reports))), InvalidOutputsWarning)
            return False

        return True

    def get_expected_reports(self, archive, sed_documents):
        """ Get the ids of the reports expected to be produced from a COMBINE/OMEX archive

        Args:
            archive (:obj:`CombineArchive`): COMBINE/OMEX archive
            sed_documents (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): dictionary that maps the
                location of each SED document to the document

        Returns:
            :obj:`set` of :obj:`str`: ids of the reports expected to be produced from a COMBINE/OMEX archive
        """
        expected_report_ids = set()
        for content in archive.contents:
            if (
                ((self._archive_has_master and content.master) or (not self._archive_has_master))
                and content.format == CombineArchiveContentFormat.SED_ML
            ):
                doc = sed_documents[content.location]
                for output in doc.outputs:
                    if isinstance(output, Report):
                        expected_report_ids.add(os.path.join(os.path.relpath(content.location, '.'), output.id))
        return expected_report_ids


class WhenACombineArchiveHasAMasterFileSimulatorOnlyExecutesThisFile(CombineArchiveTestCase):
    """ Test that when a COMBINE/OMEX archive defines a (single) master file, the simulator only
    executes this file.

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should  have a master file
    """

    def _is_concrete(): return True

    @property
    def _archive_has_master(self):
        return True


class WhenACombineArchiveHasNoMasterFileSimulatorExecutesAllSedDocuments(CombineArchiveTestCase):
    """ Test that when a COMBINE/OMEX archive does not have a defined master file, the simulator
    executes all SED-ML files.

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should  have a master file
    """

    def _is_concrete(): return True

    @property
    def _archive_has_master(self):
        return False


class CombineArchiveHasSedDocumentsInNestedDirectories(WhenACombineArchiveHasAMasterFileSimulatorOnlyExecutesThisFile):
    """ Test that SED documents in nested directories can be executed

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should  have a master file
    """

    def _is_concrete(): return True

    SED_DOCUMENT_LOCATIONS_ARE_NESTED = True
    SED_DOCUMENTS_HAVE_SAME_NAMES = False


class CombineArchiveHasSedDocumentsWithSameNamesInDifferentInNestedDirectories(
        WhenACombineArchiveHasNoMasterFileSimulatorExecutesAllSedDocuments):
    """ Test that SED documents with the same names in nested directories can be executed and their outputs are saved to distinct paths

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should  have a master file
    """

    def _is_concrete(): return True

    SED_DOCUMENT_LOCATIONS_ARE_NESTED = True
    SED_DOCUMENTS_HAVE_SAME_NAMES = True



""" Methods for checking support for reports of the execution status of modeling projects

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import Config
from ..exceptions import InvalidOutputsException, SkippedTestCaseException
from .published_project import SingleMasterSedDocumentCombineArchiveTestCase
from biosimulators_utils.combine.data_model import CombineArchive  # noqa: F401
from biosimulators_utils.config import get_config
from biosimulators_utils.log.data_model import Status
from biosimulators_utils.sedml.data_model import SedDocument, Report  # noqa: F401
import abc
import os
import requests
import requests.exceptions
import yaml

__all__ = [
    'LoggingTestCase',
    'SimulatorReportsTheStatusOfTheExecutionOfCombineArchives',
    'SimulatorReportsTheStatusOfTheExecutionOfSedDocuments',
    'SimulatorReportsTheStatusOfTheExecutionOfSedTasks',
    'SimulatorReportsTheStatusOfTheExecutionOfSedOutputs',
]


class LoggingTestCase(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that simulation tool can log its execution """

    REPORT_ERROR_AS_SKIP = False
    VALIDATE_SED_DOCUMENT_LOGS = False
    VALIDATE_TASK_LOGS = False
    VALIDATE_OUTPUT_LOGS = False

    @abc.abstractmethod
    def is_concrete(self):
        """ Whether the class is abstract

        Returns:
            :obj:`bool`: whether the class is abstract
        """
        pass  # pragma: no cover

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        log_path = os.path.join(outputs_dir, get_config().LOG_PATH)

        if not os.path.isfile(log_path):
            msg = (
                'The simulator did not export information about the status of its execution. '
                'Simulators are encouraged to stream information about their execution status.\n\n'
                'More information: https://biosimulators.org/conventions/status'
            )
            raise SkippedTestCaseException(msg)

        try:
            with open(log_path, 'r') as file:
                log = yaml.load(file, Loader=yaml.FullLoader)
        except Exception as exception:
            msg = 'The execution status report produced by the simulator is not valid:\n\n  {}'.format(
                str(exception).replace('\n', '\n  '))
            raise InvalidOutputsException(msg)

        config = Config()
        if config.runbiosimulations_api_endpoint:
            endpoint = config.runbiosimulations_api_endpoint + 'logs/validate'
            response = requests.post(endpoint, json=log)
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                msg = (
                    'The simulation log is invalid. Documentation about the log format is available at '
                    'https://biosimulators.org/conventions and https://api.biosimulators.org.'
                )
                error = response.json()['error'][0]

                pointer = error.get('source', {}).get('pointer', None)
                if pointer:
                    msg += '\n\nSource: ' + pointer

                msg += '\n\n' + error['detail']

                raise InvalidOutputsException(msg)

        self._status_valid = True

        def is_status_valid(status, self=self):
            if status not in [
                Status.SUCCEEDED.value,
                Status.SKIPPED.value,
                Status.FAILED.value,
            ]:
                self._status_valid = False

        try:
            is_status_valid(log['status'])

            if log.get('sedDocuments', None) or self.VALIDATE_SED_DOCUMENT_LOGS:
                for doc_log in log['sedDocuments']:
                    is_status_valid(doc_log['status'])

                    if doc_log.get('tasks', None) or self.VALIDATE_TASK_LOGS:
                        for task_log in doc_log['tasks']:
                            is_status_valid(task_log['status'])

                    if doc_log.get('outputs', None) or self.VALIDATE_OUTPUT_LOGS:
                        for output_log in doc_log['outputs']:
                            is_status_valid(output_log['status'])

                            els = output_log.get('dataSets', output_log.get('curves', output_log.get('surfaces', None)))
                            if els is not None:
                                for el in els:
                                    is_status_valid(el['status'])

        except Exception as exception:
            msg = 'The execution status report produced by the simulator is not valid:\n\n  {}'.format(
                str(exception).replace('\n', '\n  '))
            raise InvalidOutputsException(msg)

        if not self._status_valid:
            msg = (
                'The execution status report produced by the simulator is not valid. '
                'By the end of the execution of a COMBINE/OMEX archive, the status of '
                'the archive, each SED document, and each SED element should be '
                '`SUCCEEDED`, `SKIPPED`, or `FAILED`.'
            )
            raise InvalidOutputsException(msg)

        return True


class SimulatorReportsTheStatusOfTheExecutionOfCombineArchives(LoggingTestCase):
    """ Test that simulator logs the execution of COMBINE/OMEX archives """

    def is_concrete(self):
        """ Whether the class is concrete

        Returns:
            :obj:`bool`: whether the class is concrete
        """
        return True


class SimulatorReportsTheStatusOfTheExecutionOfSedDocuments(LoggingTestCase):
    """ Test that simulator logs the execution of individual SED document in COMBINE/OMEX archives """

    VALIDATE_SED_DOCUMENT_LOGS = True

    def is_concrete(self):
        """ Whether the class is concrete

        Returns:
            :obj:`bool`: whether the class is concrete
        """
        return True


class SimulatorReportsTheStatusOfTheExecutionOfSedTasks(LoggingTestCase):
    """ Test that simulator logs the execution of individual SED tasks in COMBINE/OMEX archives """

    VALIDATE_SED_DOCUMENT_LOGS = True
    VALIDATE_TASK_LOGS = True

    def is_concrete(self):
        """ Whether the class is concrete

        Returns:
            :obj:`bool`: whether the class is concrete
        """
        return True


class SimulatorReportsTheStatusOfTheExecutionOfSedOutputs(LoggingTestCase):
    """ Test that simulator logs the execution of individual SED outputs in COMBINE/OMEX archives """

    VALIDATE_SED_DOCUMENT_LOGS = True
    VALIDATE_OUTPUT_LOGS = True

    def is_concrete(self):
        """ Whether the class is concrete

        Returns:
            :obj:`bool`: whether the class is concrete
        """
        return True



""" Utility methods

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-06-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .config import Config
import numpy
import os

__all__ = ['get_singularity_image_filename', 'simulation_results_isnan']


def get_singularity_image_filename(docker_image):
    """ Get the location where a Singularity version of a Docker image should be saved

    Args:
        docker_image (:obj:`str`): URL for the Docker version of the image

    Returns:
        :obj:`str`: path where a Singularity version of a Docker image should be saved
    """

    return os.path.join(
        Config().singularity_image_dirname,
        docker_image.replace('/', '_').replace(':', '_') + '.sif')


def simulation_results_isnan(value):
    """ Determine whether a scalar or each element of an array is NaN.

    Args:
        value (:obj:`int`, :obj:`float`, or :obj:`numpy.ndarray`): scalar or array

    Returns:
        :obj:`bool` or :obj:`numpy.ndarray`: whether the value or each element of the value is NaN
    """
    try:
        return numpy.isnan(value)
    except TypeError:
        if isinstance(value, numpy.ndarray):
            value_type = 'a NumPy array of dtype `{}`'.format(str(value.dtype))
        else:
            value_type = 'a `{}`'.format(value.__class__.__name__)

        msg = (
            'Simulation results must be numerical (e.g., `int`, `float`, or NumPy array of dtype `float64`, `int64`). '
            'Simulation results are {}.'
        ).format(value_type)
        raise TypeError(msg)
